/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*- */
/*
 * neural-net.cc
 * Copyright (C) Scott Woodrum 2011 <swoodrum@gmail.com>
 * 
 * neural-net.cc is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * neural-net.cc is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * Derived from original Java code authored by
 * Julio César Sandria Reynoso and article at
 * http://www.javaworld.com/javaworld/jw-05-2005/jw-0516-lego.html?page=1
 */

#include "neural-net.h"
#include <math.h>

// training data
int data1a[3] = {0, 0, 0};
// expected response
int data1b[2] = {1, 1};
// training data
int data2a[3] = {1, 0, 0};
// expected response
int data2b[2] = {1, 0};
// training data
int data3a[3] = {0, 0, 1};
// expected response
int data3b[2] = {0, 1};
// training data
int data4a[3] = {0, 1, 0};
// expected response
int data4b[2] = {0, 0};
// input array
float input[4] = { 0, 0, 0, 1 };
// input layer weights
float w1[4][3] = { { 0, 0, 0 }, { 0, 0, 0 }, { 0, 0, 0 }, { 0, 0, 0 } };
// hidden layer
float hidden[3] = { 0, 0, 1 };
// hidden layer weights
float w2[3][2] = { { 0, 0 }, { 0, 0 }, { 0, 0 } };
float output[2] = { 0, 0 };
float delta2[2] = { 0, 0 };
float delta1[3] = { 0, 0, 0 };
// just a counter for number of training iterations, not used at present
int trainedEpochs = 0;

NeuralNet::NeuralNet(unsigned int randomSeed) {
	byte i, j;
	srand(randomSeed);
	// Initialize weights randomly between 0.1 and 0.9
	for (i = 0; i < 4; i++) {
		for (j = 0; j < 3; j++) {
			w1[i][j] = randomFloat() * 0.8 + 0.1;
		}
	}
	for (i = 0; i < 3; i++) {
		for (j = 0; j < 2; j++) {
			w2[i][j] = randomFloat() * 0.8 + 0.1;
		}
	}
}

// used in my debugging
float* NeuralNet::getHiddenLayer() {
	return hidden;
}

void NeuralNet::train(int e) {
	for (int i = 0; i < e; i++) {
		// Call method learn with training data
		learn(data1a, data1b);
		learn(data2a, data2b);
		learn(data3a, data3b);
		learn(data4a, data4b);
		trainedEpochs++;
	}
}

void NeuralNet::learn(int inp[], int out[]) {
	int i, j;
	double sum, out_j;
	// Initialize input units
	for (i = 0; i < 3; i++)
		input[i] = inp[i];
		
	// Calculate hidden units
	for (j = 0; j < 2; j++) {
		sum = 0;
		for (i = 0; i < 3; i++) 
			sum = sum + w1[i][j] * input[i];
			
		hidden[j] = 1 / (1 + exp(-sum));
		
	}
	
	// Calculate output units
	for (j = 0; j < 2; j++) {
		sum = 0;
		for (i = 0; i < 3; i++)
			sum = sum + w2[i][j] * hidden[i];

		output[j] = 1 / (1 + exp(-sum));
	}
	
	// Calculate delta2 errors
	for (j = 0; j < 2; j++) {
		if (out[j] == 0)
			out_j = 0.1;
		else if (out[j] == 1)
			out_j = 0.9;
		else
			out_j = out[j];
		delta2[j] = output[j] * (1 - output[j]) * (out_j - output[j]);
	}
	
// Calculate delta1 errors
	for (j = 0; j < 3; j++) {
		sum = 0;
		for (i = 0; i < 2; i++)
			sum = sum + delta2[i] * w2[j][i];

		delta1[j] = hidden[j] * (1 - hidden[j]) * sum;
	}
	
	// Adjust weights w2
	for (i = 0; i < 3; i++)
		for (j = 0; j < 2; j++)
			w2[i][j] = w2[i][j] + 0.35 * delta2[j] * hidden[i];

	// Adjust weights w1
	for (i = 0; i < 3; i++)
		for (j = 0; j < 3; j++)
			w1[i][j] = w1[i][j] + 0.35 * delta1[j] * input[i];
}

void NeuralNet::test(int inp[], int out[]) {
	int i, j;
	double sum;

	// Initialize input units
	for (i = 0; i < 3; i++)
		input[i] = inp[i];

	// Calculate hidden units
	for (j = 0; j < 2; j++) {
		sum = 0;
		for (i = 0; i < 3; i++)
			sum = sum + w1[i][j] * input[i];

		hidden[j] = 1 / (1 + exp(-sum));
	}

	// Calculate output units
	for (j = 0; j < 2; j++) {

		sum = 0;
		for (i = 0; i < 3; i++)
			sum = sum + w2[i][j] * hidden[i];

		output[j] = 1 / (1 + exp(-sum));
	}

	// Assign output to param out[]
	for (i = 0; i < 2; i++)
		if (output[i] >= 0.5)
			out[i] = 1;
		else
			out[i] = 0;
}

float NeuralNet::randomFloat() {
	return (float)rand()/(float)RAND_MAX;
}



